#include "scan.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <complex.h>

#include "phy/signal.h"
#include "utils/file.h"
#include "phy/frequency_offset.h"
#include "utils/fft.h"
#include "phy/channel.h"
#include "phy/qpsk.h"
#include "transport/fic.h"
#include "transport/fig.h"
#include "coding/frequency_interleaving.h"
#include "transport/msc.h"
#include "sdr/sdr.h"
#include "utils/io.h"
#include "audio/alsa.h"
#include "utils/log.h"
#include "utils/band3.h"

typedef struct {
  int sc_id;
  int start_address;
  int is_long;
  int option;               /* valid if is_long == 1 */
  int protection_level;     /* valid if is_long == 1 */
  int sc_size;              /* valid if is_long == 1 */
  int table_switch;         /* valid if is_long == 0 */
  int table_index;          /* valid if is_long == 0 */
  int recv_count;
} _01;

typedef struct {
  int sc_id;
  int service_reference;
  int country_id;
  int recv_count;
} _02;

typedef struct {
  int service_reference;
  int country_id;
  char charfield[17];
  int recv_count;
} _11;

typedef struct {
  _01 _01[256];
 int _01_size;
 int _01_twice;
  _02 _02[256];
 int _02_size;
 int _02_twice;
  _11 _11[256];
 int _11_size;
 int _11_twice;
} scanner_t;

static int time_sync(io *in)
{
  complex ts[T_F + T_NULL];
  int c;
  double energy;
  double min_energy;
  int min_pos;
  int i;

  c = in->get_samples(in, ts, T_F + T_NULL);
  if (c != T_F + T_NULL)
    return -1;

  energy = 0;

  for (i = 0; i < T_NULL; i++)
    energy += cabs(ts[i]);

  min_pos = 0;
  min_energy = energy;

  for (i = T_NULL; i < T_F + T_NULL; i++) {
    energy -= cabs(ts[i - T_NULL]);
    energy += cabs(ts[i]);
    if (energy < min_energy) {
      min_energy = energy;
      min_pos = i - T_NULL + 1;
    }
  }

  if (min_pos > T_F)
    min_pos -= T_F;

  in->put_back_samples(in, &ts[min_pos], T_F + T_NULL - min_pos);

  return min_pos;
}

static int fast_time_sync(io *in)
{
  complex ts[T_NULL + 200];
  int c;
  double energy;
  double min_energy;
  int min_pos;
  int i;

  c = in->get_samples(in, ts, T_NULL + 200);
  if (c != T_NULL + 200)
    return -1;

  energy = 0;

  for (i = 0; i < T_NULL; i++)
    energy += cabs(ts[i]);

  min_pos = 0;
  min_energy = energy;

  for (i = T_NULL; i < T_NULL + 200; i++) {
    energy -= cabs(ts[i - T_NULL]);
    energy += cabs(ts[i]);
    if (energy < min_energy) {
      min_energy = energy;
      min_pos = i - T_NULL + 1;
    }
  }

  if (min_pos > 200)
    min_pos -= 200;

  in->put_back_samples(in, &ts[min_pos], T_NULL + 200 - min_pos);

  return min_pos;
}

#if 0
/* return -1 if not found */
static int get_subchannel(int scid, fig0_1 *f01, int f01_size, int *sc)
{
  int i;
  int s;
  for (i = 0; i < f01_size; i++) {
    for (s = 0; s < f01[i].n; s++) {
      if (f01[i].sub_channel[s].sc_id == scid) {
        *sc = s;
        return i;
      }
    }
  }
  return -1;
}

static int get_name(int country_id, int service_ref, fig1_1 *f11, int f11_size)
{
  int i;
  for (i = 0; i < f11_size; i++) {
    if (country_id == f11[i].country_id
        && service_ref == f11[i].service_reference)
      return i;
  }
  return -1;
}
#endif

static void save_config(int chan, scanner_t *s)
{
  int i;
  int j;
  int k;

  for (i = 0; i < s->_02_size; i++) {
    /* get corresponding _01 (same sc_id) */
    for (j = 0; j < s->_01_size; j++) {
      if (s->_02[i].sc_id == s->_01[j].sc_id)
        break;
    }
    if (j == s->_01_size) {
      error("subchannel config not found, skip service %d\n",
            s->_02[i].service_reference);
      continue;
    }

    /* get name */
    for (k = 0; k < s->_11_size; k++) {
      if (s->_02[i].service_reference == s->_11[k].service_reference
          && s->_02[i].country_id == s->_11[k].country_id)
        break;
    }
    if (k == s->_11_size) {
      error("name not found, skip service %d\n",
            s->_02[i].service_reference);
      continue;
    }

    /* save */
    printf("CHAN %s [%d]%s scid %d is_long %d ",
           band3[chan].name,
           (int)strlen(s->_11[k].charfield),
           s->_11[k].charfield,
           s->_02[i].sc_id,
           s->_01[j].is_long);
    if (s->_01[j].is_long)
      printf("option %d protection_level %d sc_size %d\n",
             s->_01[j].option,
             s->_01[j].protection_level,
             s->_01[j].sc_size);
    else
      printf("table_switch %d table_index %d\n",
             s->_01[j].table_switch,
             s->_01[j].table_index);
  }

#if 0
  int i;
  int s;
  int ecc;
  int country_id;
  int service_ref;
  int sc;
  int sci;
  int t;
  int nm;

  /* get all the services that are audio with type=63 (dab+) */
  for (i = 0; i < f02_size; i++) {
    for (s = 0; s < f02[i].n; s++) {
      service_ref = f02[i].service[s].service_reference;
      country_id  = f02[i].service[s].country_id;
      ecc         = f02[i].service[s].ecc;
      (void)ecc;
      for (t = 0; t < f02[i].service[s].n; t++) {
        if (f02[i].service[s].tm[t].ps == 0) {
          warn("skip non primary service\n");
          continue;
        }
        if (f02[i].service[s].tm[t].tm_id != 0) {
          warn("skip non audio service\n");
          continue;
        }
        if (f02[i].service[s].tm[t].asct != 63) {
          warn("skip non DAB+ audio service\n");
          continue;
        }
        if (f02[i].service[s].tm[t].ca_flag != 0) {
          warn("skip DAB+ audio service with Access Control\n");
          continue;
        }
        /* get subchannel parameters from fig01 (if present) */
        sc = get_subchannel(f02[i].service[s].tm[t].scid, f01, f01_size, &sci);
        if (sc == -1) {
          warn("skip DAB+ audio service, no subchannel found\n");
          continue;
        }
        /* get name */
        nm = get_name(country_id, service_ref, f11, f11_size);
        if (nm == -1) {
          warn("skip DAB+ audio service, no name found\n");
          continue;
        }
        /* dump config */
        printf("CHAN %s [%d]%s scid %d is_long %d ",
               band3[chan].name,
               (int)strlen(f11[nm].header.charfield),
               f11[nm].header.charfield,
               f02[i].service[s].tm[t].scid,
               f01[sc].sub_channel[sci].is_long);
        if (f01[sc].sub_channel[sci].is_long)
          printf("option %d protection level %d sc_size %d\n",
                 f01[sc].sub_channel[sci].option,
                 f01[sc].sub_channel[sci].protection_level,
                 f01[sc].sub_channel[sci].sc_size);
        else
          printf("table_switch %d table_index %d\n",
                 f01[sc].sub_channel[sci].table_switch,
                 f01[sc].sub_channel[sci].table_index);
      }
    }
  }
#endif
}

void process_f01(scanner_t *s, fig0_1 *f)
{
  int i;
  int j;
  for (j = 0; j < f->n; j++) {
    for (i = 0; i < s->_01_size; i++) {
      if (s->_01[i].sc_id == f->sub_channel[j].sc_id)
        break;
    }
    /* if already there, increase recv_count and update _twice if needed */
    if (i != s->_01_size) {
      s->_01[i].recv_count++;
      if (s->_01[i].recv_count == 2)
        s->_01_twice++;
      continue;
    }
    /* not there, insert if free space */
    if (s->_01_size == 256) { error("f01 full\n"); exit(1); }
    s->_01_size++;
    s->_01[i].sc_id            = f->sub_channel[j].sc_id;
    s->_01[i].start_address    = f->sub_channel[j].start_address;
    s->_01[i].is_long          = f->sub_channel[j].is_long;
    s->_01[i].option           = f->sub_channel[j].option;
    s->_01[i].protection_level = f->sub_channel[j].protection_level;
    s->_01[i].sc_size          = f->sub_channel[j].sc_size;
    s->_01[i].table_switch     = f->sub_channel[j].table_switch;
    s->_01[i].table_index      = f->sub_channel[j].table_index;
    s->_01[i].recv_count = 1;
  }
}

void process_f02(scanner_t *s, fig0_2 *f)
{
  int i, j, k;

  for (j = 0; j < f->n; j++) {
    for (k = 0; k < f->service[j].n; k++) {
      if (f->service[j].tm[k].ps == 0) {
        warn("skip non primary service\n");
        continue;
      }
      if (f->service[j].tm[k].tm_id != 0) {
        warn("skip non audio service\n");
        continue;
      }
      if (f->service[j].tm[k].asct != 63) {
        warn("skip non DAB+ audio service\n");
        continue;
      }
      if (f->service[j].tm[k].ca_flag != 0) {
        warn("skip DAB+ audio service with Access Control\n");
        continue;
      }
      for (i = 0; i < s->_02_size; i++) {
        if (s->_02[i].service_reference == f->service[j].service_reference
            && s->_02[i].country_id == f->service[j].country_id)
          break;
      }
      /* if already there, increase recv_count and update _twice if needed */
      if (i != s->_02_size) {
        s->_02[i].recv_count++;
        if (s->_02[i].recv_count == 2)
          s->_02_twice++;
        continue;
      }
      /* not there, insert if free space */
      if (s->_02_size == 256) { error("f02 full\n"); exit(1); }
      s->_02_size++;
      s->_02[i].sc_id             = f->service[j].tm[k].scid;
      s->_02[i].service_reference = f->service[j].service_reference;
      s->_02[i].country_id        = f->service[j].country_id;
      s->_02[i].recv_count = 1;
    }
  }
}

void process_f11(scanner_t *s, fig1_1 *f)
{
  int i;

  for (i = 0; i < s->_11_size; i++) {
    if (s->_11[i].service_reference == f->service_reference
        && s->_11[i].country_id == f->country_id)
      break;
  }
  if (i != s->_11_size) {
    s->_11[i].recv_count++;
    if (s->_11[i].recv_count == 2)
      s->_11_twice++;
    return;
  }
  if (s->_11_size == 256) { error("f11 full\n"); exit(1); }
  s->_11_size++;
  s->_11[i].service_reference = f->service_reference;
  s->_11[i].country_id = f->country_id;
  memcpy(s->_11[i].charfield, f->header.charfield, 17);
  s->_11[i].recv_count = 1;
}

void reset_scanner(scanner_t *s)
{
  s->_01_size = s->_01_twice = 0;
  s->_02_size = s->_02_twice = 0;
  s->_11_size = s->_11_twice = 0;
}

void scan(void)
{
  io *in;
  int i;
  int j;
  int start;
  double fine_freq_offset;
  complex *in_time;
  complex *_cur_freq;
  complex *_prev_freq;
  complex *cur_freq;
  complex *prev_freq;
  complex *channel;
  complex *z_ref;
  fft_data *fft_data;
  int freq_offset_cur_pos;
  complex qpsk[K];
  int p[K*2];
  fic *fic;
  frequency_interleaver *fl;
  frame_figs figs;
  int frame_cif_id;
  int prev_frame_cif_id = -1;
  int fig0_0_errors;
  int fig0_0_present;
  scanner_t s;
  int dont_process_fig;

  double start_frequency;

  fic = new_fic();

  fl = new_frequency_interleaver();

  init_frequency_offset();

  in_time = malloc(sizeof(complex) * T_S);
  _cur_freq = malloc(sizeof(complex) * T_U);
  _prev_freq = malloc(sizeof(complex) * T_U);
  channel = malloc(sizeof(complex) * T_U);
  z_ref = malloc(sizeof(complex) * T_U);
  if (in_time == NULL || _cur_freq == NULL || _prev_freq == NULL
      || channel == NULL || z_ref == NULL) abort();

  cur_freq = _cur_freq;
  prev_freq = _prev_freq;

  fft_data = init_fft(T_U);

  compute_phase_ref_full(fft_data, z_ref);

  in = new_sdr("usrp");
  if (in == NULL) exit(1);

  int fff = -1;

  in->set_gain(in, 1);
  in->set_freq(in, 10000000); /* dumb value (10MHz) for start */
  in->start(in);

next_freq:
  fff++;
  if (fff == band3_size) goto over;
  printf("digging %s (%g)\n", band3[fff].name, band3[fff].freq);
  start_frequency = band3[fff].freq * 1000000;
  fig0_0_present = 0;
//start_frequency = 225648000;
printf("freq %g %g\n", start_frequency, band3[fff].freq * 1000000);
  in->reset(in);
  in->set_freq(in, start_frequency); //225648000); //220352000); //201072000); //195936000); //216928000);

restart:
printf("RESET LOOOPPP\n");
  fig0_0_errors = 0;
  reset_scanner(&s);

  start = time_sync(in);
  if (start == -1) { printf("no time sync\n"); goto next_freq; }

next_frame:
  debug("#start %d\n", start);

  if (in->get_samples(in, NULL, T_NULL) != T_NULL) goto over;

  figs.size = 0;

  /* symbol 1 */
  if (in->get_samples(in, in_time, T_S) != T_S) goto over;
  fine_freq_offset = frequency_offset_fine(in_time);
  debug("# %g ", fine_freq_offset);
  freq_offset_cur_pos = remove_frequency_offset(0, 
                            in_time, fine_freq_offset);
  fine_freq_offset = frequency_offset_fine(in_time);
  debug("%g\n", fine_freq_offset);

{
int i;
double e = 0;
for (i= 0; i < T_S; i++) e += cabs(in_time[i]);
printf("energy %g\n", e / T_S);
}

  fft(fft_data, &in_time[Delta], prev_freq);

  dont_process_fig = 0;
  for (i = 2; i < 5; i++) {
    if (in->get_samples(in, in_time, T_S) != T_S) goto over;
    fine_freq_offset = frequency_offset_fine(in_time);
    debug("#%g ", fine_freq_offset);
    freq_offset_cur_pos = remove_frequency_offset(freq_offset_cur_pos,
                              in_time, fine_freq_offset);
    fine_freq_offset = frequency_offset_fine(in_time);
    debug("%g\n", fine_freq_offset);

    fft(fft_data, &in_time[Delta], cur_freq);

    /* frequency de-interleaving done here (using k2n) */
    /* positive frequencies [1..768] */
    for (j = 0; j < K/2; j++)
      qpsk[fl->k2n[FK(j+1)]] = cur_freq[j+1] * conj(prev_freq[j+1]);
    /* negative frequencies [-768..-1] */
    for (j = 0; j < K/2; j++)
      qpsk[fl->k2n[FK(j-K/2)]] = cur_freq[j+T_U-K/2] * conj(prev_freq[j+T_U-K/2]);

    symbol_unmap(p, qpsk);

    for (j = 0; j < 4; j++) {
      if (receive_fic_bits(&figs, fic, &p[768*j], (i-2)*4+j)) {
        /* error, clear all */
        warn("some FIC not received correctly, reset scan\n");
        reset_scanner(&s);
        dont_process_fig = 1;
      }
    }

#define  SWAP(a, b) do { complex *x = a; a = b; b = x; } while (0)
    SWAP(cur_freq, prev_freq);
  }

  int f;
  if (dont_process_fig == 0) for (f = 0; f < figs.size; f++) {
    switch (figs.f[f].type) {
    default: break;
    case FIG0_1: process_f01(&s, &figs.f[f].u.fig0_1); break;
    case FIG0_2: process_f02(&s, &figs.f[f].u.fig0_2); break;
    case FIG1_1: process_f11(&s, &figs.f[f].u.fig1_1); break;
    }
  }

  /* check that we have FIG 0/0 */
  if (figs.size) printf("fig\n");
  if (figs.size == 0 || figs.f[0].type != FIG0_0) {
if (figs.size) printf("type %d size %d\n", figs.f[0].type, figs.size);
    debug("error: FIG 0/0 not received, skip frame\n");
    fig0_0_errors++;
    if (fig0_0_errors == 10) {
      debug("error: too much missed FIG 0/0, full resync\n");
      if (fig0_0_present) goto restart;
      goto next_freq;
    }
    goto frame_over;
  }
  fig0_0_present++;
  /* we got it all twice? next freq */
  if (s._01_size && s._01_size == s._01_twice
      && s._02_size && s._02_size == s._02_twice
      && s._11_size && s._11_size == s._11_twice) {
    save_config(fff, &s);
    goto next_freq;
  }
  fig0_0_errors = 0;
  frame_cif_id = figs.f[0].u.fig0_0.cif_count;
  if (frame_cif_id != prev_frame_cif_id + 4) {
    warn("some missing frames detected, reset scan\n");
    reset_scanner(&s);
  }
  prev_frame_cif_id = frame_cif_id;

frame_over:
  /* skip msc symbols */
printf("i %d\n", i);
  for (; i < 77; i++) {
    if (in->get_samples(in, in_time, T_S) != T_S) goto over;
  }

  /* put back last 100 samples of last symbol, for fast time sync */
  in->put_back_samples(in, &in_time[T_S-100], 100);
  start = fast_time_sync(in);
  if (start == -1) goto restart;

  goto next_frame;

over:
  in->stop(in);
  in->free(in);
}
